import os
from pathlib import Path

import click

from vertagus.configuration import load
from vertagus.configuration import types as cfgtypes
from vertagus import factory
from vertagus import operations as ops


@click.command(name="create-tag")
@click.option(
    "--config", 
    "-c", 
    default=str(Path(os.getcwd()) / "vertagus.toml"), 
    help="Path to the configuration file"
)
@click.option(
    "--stage-name",
    "-s",
    default=None,
    help="Name of a stage"
)
@click.option(
    "--ref",
    "-r",
    default=None,
    help="An SCM ref that should be tagged. Default is current commit."
)
def create_tag_cmd(config, stage_name, ref):
    master_config = load.load_config(config)
    scm = factory.create_scm(
        data=cfgtypes.ScmData(**master_config["scm"])
    )
    default_package_root = Path(config).parent
    if "root" not in master_config["project"]:
        master_config["project"]["root"] = default_package_root
    project = factory.create_project(
        cfgtypes.ProjectData.from_project_config(master_config["project"]),
    )
    return ops.create_tags(
        scm=scm,
        project=project,
        stage_name=stage_name,
        ref=ref
    )



from vertagus.core.manifest_base import ManifestBase
import yaml
import os.path

class YamlManifest(ManifestBase):
    manifest_type: str = "yaml"

    def __init__(self,
                 name: str,
                 path: str,
                 loc: list = None,
                 root: str = None
                 ):
        super().__init__(name, path, loc, root)
        self._doc = self._load_doc()

    @property
    def version(self):
        if not self.loc:
            raise ValueError(f"No loc provided for manifest {self.name!r}")
        p = self._doc
        for k in self.loc:
            if k not in p:
                raise ValueError(
                    f"Invalid loc {self.loc!r} for manifest {self.name!r}. "
                    f"Key {k!r} not found in {list(p.keys())}"
                )
            p = p[k]
        return p

    def _load_doc(self):
        path = self._full_path()
        with open(path) as f:
            return yaml.safe_load(f)

    def _full_path(self):
        path = self.path
        if self.root:
            path = os.path.join(self.root, path)
        return path



import click

from vertagus.rules.comparison.loader import (
    get_rules as get_comparison_rules,
    VersionComparisonRule
)
from vertagus.rules.single_version.loader import (
    get_rules as get_single_version_rules,
    SingleVersionRule
)
from vertagus.cli.formatting import DisplayTableFormatter

@click.command('list-rules')
def list_rules_cmd():
    single_version_rules: list[type[SingleVersionRule]] = get_single_version_rules()
    comparison_rules: list[type[VersionComparisonRule]] = [
        r for r in get_comparison_rules()
        if r.name != "manifests_comparison"
    ]
    formatter = DisplayTableFormatter(max_width=240)
    header = ('Rule Name', 'Config Usage', 'Description')
    rules_rows = [header]
    for rule in comparison_rules:
        rules_rows.append((rule.name, 'increment', rule.description))
    for rule in single_version_rules:
        rules_rows.append((rule.name, 'current', rule.description))
    formatter.write_table(
        rules_rows,
        col_widths=[22, 22, 100],
        header=True
    )
    click.echo()
    click.echo(formatter.getvalue())
    click.echo()



import typing as T
from vertagus.core.manifest_base import ManifestBase
from vertagus.core.rule_bases import SingleVersionRule, VersionComparisonRule


class Package:

    def __init__(self,
                 manifests: list[ManifestBase],
                 current_version_rules: list[T.Type[SingleVersionRule]],
                 version_increment_rules: list[T.Type[VersionComparisonRule]],
                 manifest_versions_comparison_rules: list[T.Type[VersionComparisonRule]]
                 ):
        self._manifests = manifests or []
        self._current_version_rules = current_version_rules or []
        self._version_increment_rules = version_increment_rules or []
        self._manifest_versions_comparison_rules = manifest_versions_comparison_rules or []



from logging import getLogger
import yaml
import tomli
from .types import MasterConfig
from vertagus.utils.config import is_yaml, is_toml


logger = getLogger(__name__)


def load_config(filepath: str) -> MasterConfig:
    logger.info(f"Loading configuration from {filepath}")
    with open(filepath, "rt") as f:
        doc = f.read()

    if is_yaml(doc):
        return yaml.safe_load(doc)
    elif is_toml(doc):
        return tomli.loads(doc)
    else:
        raise ValueError(
            "Invalid configuration file format. Supported formats are YAML and TOML. "
            "If you are attempting to load one of these file types, you can receive a "
            "more detailed error message by ensuring that your configuration file uses "
            "the correct file extension."
        )



import typing as T
from .setuptools_ import SetuptoolsPyprojectManifest
from .json_manifest import JsonManifest
from .yaml_manifest import YamlManifest

_manifest_types = {
    SetuptoolsPyprojectManifest.manifest_type: SetuptoolsPyprojectManifest,
    JsonManifest.manifest_type: JsonManifest,
    YamlManifest.manifest_type: YamlManifest,
}


def get_manifest_cls(manifest_type: str) -> T.Type[SetuptoolsPyprojectManifest]:
    if manifest_type not in _manifest_types:
        raise ValueError(f"Unknown manifest type: {manifest_type}")
    return _manifest_types[manifest_type]


def register_manifest_cls(manifest_cls: T.Type[SetuptoolsPyprojectManifest]):
    _manifest_types[manifest_cls.manifest_type] = manifest_cls



class Tag:

    def __init__(self, tag_text: str):
        self.tag_text = tag_text

    def as_string(self, prefix: str = None) -> str:
        if prefix:
            return prefix + self.tag_text
        return self.tag_text


class AliasBase(Tag):
    name: str = ""



import logging
import os

logging.basicConfig(
    level=os.environ.get("VERTAGUS_LOG_LEVEL", "INFO"),
    format="{message}",
    style="{"
)


import click
from .commands import (
    validate_cmd,
    create_tag_cmd,
    create_aliases_cmd,
    list_rules_cmd
)


@click.group()
def cli():
    pass


cli.add_command(validate_cmd)
cli.add_command(create_tag_cmd)
cli.add_command(create_aliases_cmd)
cli.add_command(list_rules_cmd)



import typing as T
from dataclasses import dataclass, field, asdict
import os

class ScmConfigBase(T.TypedDict):
    scm_type: str


ScmConfig = T.Union[ScmConfigBase, dict]


class ProjectConfig(T.TypedDict):
    manifests: list[T.Type["ManifestConfig"]]
    rules: "RulesConfig"
    stages: dict[str, "StageConfig"]
    aliases: T.Optional[list[str]]
    root: T.Optional[str]


class ManifestConfig(T.TypedDict):
    type: str
    path: str
    loc: T.Optional[str]


class ManifestComparisonConfig(T.TypedDict):
    manifests: list[str]


class RulesConfig(T.TypedDict):
    current: list[str]
    increment: list[str]
    manifest_comparisons: list[ManifestComparisonConfig]


class StageConfig(T.TypedDict):
    name: str
    manifests: T.Optional[list[str]]
    rules: T.Optional["RulesConfig"]
    aliases: T.Optional[list[str]]


class MasterConfig(T.TypedDict):
    project: ProjectConfig
    scm: T.Union[ScmConfigBase, dict]


@dataclass
class RulesData:
    current: list[str] = field(default_factory=list)
    increment: list[str] = field(default_factory=list)
    manifest_comparisons: list[ManifestComparisonConfig] = field(default_factory=list)


@dataclass
class ManifestData:
    name: str
    type: str
    path: str
    loc: str = field(default=None)

    def __init__(self, name: str, type: str, path: str, loc: str = None):
        self.name = name
        self.type = type
        self.path = path
        self.loc = self._parse_loc(loc)

    def _parse_loc(self, loc):
        if isinstance(loc, str):
            return loc.split(".")
        return loc

    def config(self):
        return dict(name=self.name, path=self.path, loc=self.loc)



class StageData:

    def __init__(self,
                 name: str,
                 manifests: list[ManifestData],
                 rules: RulesData,
                 aliases: list[str] = None
                 ):
        self.name: str = name
        self.manifests: list[ManifestData] = manifests
        self.rules: RulesData = rules
        self.aliases: list[str] = aliases

    @classmethod
    def from_stage_config(cls, name: str, config: StageConfig):
        return cls(
            name=name,
            manifests=[ManifestData(**m) for m in config.get("manifests", [])],
            rules=RulesData(
                current=config.get("rules", {}).get("current", []),
                increment=config.get("rules").get("increment", []),
                manifest_comparisons=config.get("rules").get("manifest_comparisons", []),
            ),
            aliases=config.get("aliases", []),
        )

    def config(self):
        return dict(
            name=self.name,
            manifests=[m.config() for m in self.manifests],
            current_version_rules=self.rules.current,
            version_increment_rules=self.rules.increment,
            manifest_versions_comparison_rules=self.rules.manifest_comparisons,
            aliases=self.aliases,
        )


class ProjectData:
    
    def __init__(self,
                 manifests: list[ManifestData],
                 rules: RulesData,
                 stages: list[StageData] = None,
                 aliases: list[str] = None,
                 root: str = None 
                 ):
        self.manifests: list[ManifestData] = manifests
        self.rules: RulesData = rules
        self.stages: list[StageData] = stages
        self.aliases: list[str] = aliases
        self.root: str = root or os.getcwd()

    def config(self):
        return dict(
            manifests=[m.config() for m in self.manifests],
            stages = [stage.config() for stage in self.stages.values()],
            current_version_rules=self.rules.current,
            version_increment_rules=self.rules.increment,
            manifest_versions_comparison_rules=self.rules.manifest_comparisons,
            aliases=self.aliases,
        )
    
    @classmethod
    def from_project_config(cls, config: ProjectConfig):
        stages = config.get("stages", {})
        return cls(
            manifests=[ManifestData(**m) for m in config.get("manifests", [])],
            rules=RulesData(
                current=config.get("rules", {}).get("current", []),
                increment=config.get("rules").get("increment", []),
                manifest_comparisons=config.get("rules").get("manifest_comparisons", []),
            ),
            stages=[StageData.from_stage_config(name, data) for name, data in stages.items()],
            aliases=config.get("aliases", []),
            root=config.get("root", None)
        )


class ScmData:
    
    def __init__(self, type: str, root: str = None, **kwargs):
        self.scm_type = type
        self.root = root
        self.kwargs = kwargs

    def config(self):
        return dict(
            root=self.root,
            **self.kwargs
        )



import typing as T
from .git_ import GitScm
from vertagus.core.scm_base import ScmBase

_scm_types = {
    GitScm.scm_type: GitScm,
}


def get_scm_cls(scm_type: str) -> T.Type[ScmBase]:
    if scm_type not in _scm_types:
        raise ValueError(f"Unknown scm type: {scm_type}")
    return _scm_types[scm_type]


def register_scm_cls(scm_cls: T.Type[ScmBase]):
    _scm_types[scm_cls.scm_type] = scm_cls



import typing as T
from . import library
from vertagus.core.tag_base import AliasBase


def load_aliases() -> list[T.Type[AliasBase]]:
    _rules = []
    for objname in dir(library):
        maybeobj = getattr(library, objname)
        if isinstance(maybeobj, type) and issubclass(maybeobj, AliasBase):
            obj: AliasBase = maybeobj
            if obj.name and obj.name != "base":
                _rules.append(obj)
    return _rules


def get_aliases(alias_names=None) -> list[T.Type[AliasBase]]:
    if not alias_names:
        return []
    aliases: list[T.Type[AliasBase]] = load_aliases() or []
    not_found = set(alias_names) - {alias.name for alias in aliases}
    if not_found:
        raise ValueError(f"Aliases not found: {not_found}")
    alias_d = {alias.name: alias for alias in aliases if alias.name in alias_names}
    return [alias_d[alias_name] for alias_name in alias_names]



import re
from vertagus.core.rule_bases import SingleVersionRule
from vertagus.utils import regex as regex_utils


class classproperty(object):

    def __init__(self, f):
        self.f = f
    
    def __get__(self, obj, owner):
        return self.f(owner)


class NotEmpty(SingleVersionRule):
    name = "not_empty"
    description = "Version must not be empty."

    @classmethod
    def validate_version(cls, version):
        return bool(version)
    

class RegexRuleBase(SingleVersionRule):
    pattern: str = ""

    @classproperty
    def description(cls):
        return f"Version must match the pattern: {cls.pattern}"

    @classmethod
    def validate_version(cls, version):
        return bool(re.match(cls.pattern, version))


# Major-Minor-Patch Regex Rules

class RegexMmp(RegexRuleBase):
    name = "regex_mmp"
    pattern = regex_utils.patterns["mmp"]


class RegexDevMmp(RegexRuleBase):
    name = "regex_dev_mmp"
    pattern = regex_utils.patterns["dev_mmp"]


class RegexBetaMmp(RegexRuleBase):
    name = "regex_beta_mmp"
    pattern = regex_utils.patterns["beta_mmp"]


class RegexRcMmp(RegexRuleBase):
    name = "regex_rc_mmp"
    pattern = regex_utils.patterns["rc_mmp"]


class RegexAlphaMmp(RegexRuleBase):
    name = "regex_alpha_mmp"
    pattern = regex_utils.patterns["alpha_mmp"]


# Major-Minor Regex Rules

class RegexMm(RegexRuleBase):
    name = "regex_mm"
    pattern = regex_utils.patterns["mm"]


class RegexDevMm(RegexRuleBase):
    name = "regex_dev_mm"
    pattern = regex_utils.patterns["dev_mm"]


class RegexBetaMm(RegexRuleBase):
    name = "regex_beta_mm"
    pattern = regex_utils.patterns["beta_mm"]


class RegexRcMm(RegexRuleBase):
    name = "regex_rc_mm"
    pattern = regex_utils.patterns["rc_mm"]


class RegexAlphaMm(RegexRuleBase):
    name = "regex_alpha_mm"
    pattern = regex_utils.patterns["alpha_mm"]




import typing as T


class ManifestBase:
    manifest_type: str = "base"
    version: str
    loc: list[T.Union[str, int]] = []

    def __init__(self,
                 name: str,
                 path: str,
                 loc: list[T.Union[str, int]],
                 root: str = None
                 ):
        self.name = name
        self.path = path
        if loc:
            self.loc = loc
        self.root = root

    def __repr__(self):
        return f"{self.__class__.__name__}({self.name}, {self.path}, {self.loc}, {self.version})"



from vertagus.core.tag_base import AliasBase


class StringAlias(AliasBase):
    name = ""
    alias_value = ""
    use_prefix = False

    def as_string(self, prefix: str = None) -> str:
        prefix = prefix or ""
        if self.use_prefix:
            return prefix + self.alias_value
        return self.alias_value


class StableAlias(StringAlias):
    name = "string:stable"
    alias_value ="stable"


class LatestAlias(StringAlias):
    name = "string:latest"
    alias_value ="latest"


class StablePrefixedAlias(StringAlias):
    name = "string:prefixed:stable"
    alias_value ="stable"
    use_prefix = True


class LatestPrefixedAlias(StringAlias):
    name = "string:prefixed:latest"
    alias_value ="latest"
    use_prefix = True


class MajorMinor(AliasBase):
    name = "major.minor"

    def as_string(self, prefix: str = None) -> str:
        prefix = prefix or ""
        parts = self.tag_text.split(".")
        if len(parts) < 2:
            raise ValueError(
                f"Version must have at least two parts. Found version "
                f"{self.tag_text}"
            )
        return prefix + ".".join(parts[:2])



from logging import getLogger
import yaml
import tomli

logger = getLogger(__name__)


def is_yaml(doc, filepath: str = None) -> bool:
    try:
        yaml.safe_load(doc)
        return True
    except yaml.YAMLError as e:
        if filepath and any([filepath.endswith(ext) for ext in [".yaml", ".yml"]]):
            raise e
        return False


def is_toml(doc: str, filepath: str = None) -> bool:
    try:
        tomli.loads(doc)
        return True
    except tomli.TOMLDecodeError as e:
        if filepath and filepath.endswith(".toml"):
            raise e
        return False
    



from vertagus.core.manifest_base import ManifestBase
from ..toml_manifest import TomlManifest


class SetuptoolsPyprojectManifest(TomlManifest):
    manifest_type: str = "setuptools_pyproject"
    loc = ["project", "version"]

    def __init__(self,
                 name: str,
                 path: str,
                 loc: list = None,
                 root: str = None
                 ):
        super().__init__(name, path, loc, root)
        if loc:
            self.loc = loc
        self._doc = self._load_doc()



import typing as T
from . import library
from vertagus.core.rule_bases import VersionComparisonRule


def load_rules():
    _rules = []
    for objname in dir(library):
        maybeobj = getattr(library, objname)
        if isinstance(maybeobj, type) and issubclass(maybeobj, VersionComparisonRule):
            obj: T.Type[VersionComparisonRule] = maybeobj
            if obj.name != "base":
                _rules.append(obj)
    return _rules


def get_rules(rule_names=None) -> list[T.Type[VersionComparisonRule]]:
    rules = load_rules()
    rules: list[T.Type[VersionComparisonRule]] = rules
    if rule_names is None:
        rule_names = [rule.name for rule in rules]
    not_found = set(rule_names) - {rule.name for rule in rules}
    if not_found:
        raise ValueError(f"Rules not found: {not_found}")
    rules_d = {rule.name: rule for rule in rules if rule.name in rule_names}
    return [rules_d[rule_name] for rule_name in rule_names]



from .tag_base import Tag
import typing as T


class ScmBase:
    
    scm_tpe = "base"
    tag_prefix: T.Optional[str] = None

    def __init__(self, root: str, **kwargs):
        raise NotImplementedError()

    def create_tag(self, tag: Tag, ref: str=None):
        raise NotImplementedError()
    
    def delete_tag(self, tag_name: str):
        raise NotImplementedError()
    
    def list_tags(self, prefix: str=None):
        raise NotImplementedError()

    def get_highest_version(self, prefix: str=None):
        raise NotImplementedError()

    def migrate_alias(self, alias: str, ref: str = None):
        raise NotImplementedError()



import typing as T
import json


class Rule:
    name: str = "base"
    description: str = "Base class for all rules"


class VersionComparisonRule(Rule):

    def __init__(self, config: dict):
        self.config = config

    def validate_comparison(self, versions: T.Sequence[str]):
        raise NotImplementedError(
            'Method `validate_comparison` must be implemented in subclass'
        )
    
    def __hash__(self):
        return hash(json.dumps(self.__dict__, default=str, sort_keys=True))


class SingleVersionRule(Rule):

    @classmethod
    def validate_version(cls, version: str):
        raise NotImplementedError(
            'Method validate must be implemented in subclass'
        )



import os
from logging import getLogger

import git
from git.remote import Remote
from git.exc import GitCommandError
from packaging.version import parse as parse_version, InvalidVersion
from vertagus.core.scm_base import ScmBase
from vertagus.core.tag_base import Tag, AliasBase

logger = getLogger(__name__)


class GitScm(ScmBase):

    scm_type = "git"
    _default_user_data = {
        "name": "vertagus",
        "email": "vertagus@none"
    }
    _default_remote_name = "origin"

    def __init__(self,
                 root: str = None,
                 tag_prefix: str = None,
                 user_data: dict = None,
                 remote_name: str = None
                 ):
        self.root = root or os.getcwd()
        self.tag_prefix = tag_prefix
        self.user_data = user_data or self._default_user_data
        self.remote_name = remote_name or self._default_remote_name
        self._repo = self._initialize_repo()

    @property
    def remote(self) -> Remote:
        return self._repo.remotes[self.remote_name]


    def create_tag(self, tag: Tag, ref: str=None):
        tag_prefix = self.tag_prefix or ""
        tag_text = tag.as_string(tag_prefix)
        if ref:
            commit = self._repo.commit(ref)
        else:
            commit = self._repo.head.commit
        logger.info(
            f"Creating tag {tag_text} at commit {commit}"
        )
        self._repo.create_tag(
            path=tag_text,
            ref=commit,
            message=tag_text,
        )
        self._repo.git.push(tags=True)
    
    def delete_tag(self, tag: Tag):
        _tags = [t.name for t in self._repo.tags]
        logger.debug(
            f"Tags found: {_tags}"
        )
        tag_text = tag.as_string(self.tag_prefix)
        try:
            self._repo.delete_tag(tag_text)
        except GitCommandError as e:
            logger.warning(
                f"Error encountered while deleting local tag {tag_text!r}: {e.__class__.__name__}: {e}"
            )
        try:
            self._repo.git.execute(["git", "push", "--delete", self.remote_name, tag_text])
        except GitCommandError as e:
            logger.warning(
                f"Error encountered while deleting remote tag {tag_text!r}: {e.__class__.__name__}: {e}"
            )
        self._repo.git.push(tags=True)
    
    def list_tags(self, prefix: str=None):
        tags = [
            t.split("tags/")[-1].strip() for t in
            self._repo.git.execute(["git", "ls-remote", "--tags", self.remote_name]).split("\n")
            if not t.endswith("^{}")
        ]
        if not prefix and self.tag_prefix:
            prefix = self.tag_prefix
        if prefix:
            tags = [tag for tag in tags if tag.startswith(prefix)]
        return tags

    def migrate_alias(self, alias: AliasBase, ref: str = None):
        logger.info(
            f"Migrating alias {alias.name} to ref {ref}"
        )
        try:
            self.delete_tag(alias)
        except GitCommandError as e:
            logger.error(f"Error encountered while deleting alias {alias.name}: {e.__class__.__name__}: {e}")
        self.create_tag(alias, ref=ref)

    def get_highest_version(self, prefix: str = None):
        if not prefix and self.tag_prefix:
            prefix = self.tag_prefix
        tags = self.list_tags(prefix=prefix)
        if not tags:
            return None
        versions = tags
        if prefix:
            versions = [tag.replace(prefix, "") for tag in tags]
        
        valid_versions = []
        for version in versions:
            try:
                parse_version(version)
                valid_versions.append(version)
            except InvalidVersion:
                logger.warning(f"Invalid version found: {version}")
        if not valid_versions:
            return None
        return max(valid_versions, key=lambda v: parse_version(v))
    
    def _initialize_repo(self):
        repo = git.Repo(self.root)
        logger.debug(
            f"Initializing git repository at {self.root} "
            f"with user data {self.user_data}."
        )
        repo.config_writer().set_value(
            "user", "name", self.user_data['name']
        ).release()
        repo.config_writer().set_value(
            "user", "email", self.user_data['email']
        ).release()
        return repo



import typing as T
from logging import getLogger

from vertagus.core.manifest_base import ManifestBase
from vertagus.core.rule_bases import SingleVersionRule, VersionComparisonRule
from vertagus.rules.comparison.library import ManifestsComparisonRule
from vertagus.core.tag_base import AliasBase
from .package_base import Package
from .stage import Stage


logger = getLogger(__name__)


class Project(Package):
    
    def __init__(self,
                 manifests: list[ManifestBase],
                 current_version_rules: list[T.Type[SingleVersionRule]],
                 version_increment_rules: list[VersionComparisonRule],
                 manifest_versions_comparison_rules: list[ManifestsComparisonRule],
                 stages: list[Stage] = None,
                 aliases: list[type[AliasBase]] = None
                 ):
        super().__init__(
            manifests=manifests,
            current_version_rules=current_version_rules,
            version_increment_rules=version_increment_rules,
            manifest_versions_comparison_rules=manifest_versions_comparison_rules
        )
        self._stages = stages or []
        self.aliases = aliases or []
    @property
    def stages(self):
        return self._stages
    
    def get_version(self, stage_name: str = None):
        manifests = self._get_manifests(stage_name)
        if not manifests:
            raise ValueError("No manifests found.")
        return manifests[0].version

    def get_aliases(self, stage_name: str = None) -> list[AliasBase]:
        version = self.get_version()
        aliases = self._get_version_aliases(version)
        if stage_name:
            stage = self._get_stage(stage_name)
            aliases.extend(stage.get_version_aliases(version))
        return list(dict.fromkeys(aliases).keys())

    def validate_version(self, previous_version: str, stage_name: str = None):
        primary_manifest = self._get_manifests(stage_name)[0]
        current_version = self.get_version(stage_name)
        logger.info(
            f"Validating version for manifest {primary_manifest.name!r}\n"
            f"  manifest_type={primary_manifest.manifest_type!r}\n"
            f"  path={primary_manifest.path}\n"
            f"  version loc={primary_manifest.loc}\n"
            f"  previous_version={previous_version}\n"
            f"  current_version={current_version}\n"
            f"  stage_name={stage_name}"
        )
        validated = self._run_current_version_rules(current_version, stage_name)
        if not validated:
            return validated
        validated = self._run_version_increment_rules(previous_version, current_version, stage_name)
        if not validated:
            return validated
        return self._run_manifest_versions_comparison_rules(stage_name)

    def _get_version_aliases(self, version: str) -> list[AliasBase]:
        return [alias(version) for alias in self.aliases]

    def _run_current_version_rules(self, current_version, stage_name=None):
        validated = True
        for rule in self._get_current_version_rules(stage_name):
            logger.info(
                f"Validating rule {rule.name!r} for {current_version}"
            )
            validated = rule.validate_version(current_version)
            if not validated:
                logger.error(
                    f"Validation failed for rule {rule.name!r}: {rule.description}"
                )
                return validated
        return validated

    def _run_version_increment_rules(self, previous_version, current_version, stage_name=None):
        validated = True
        versions = [previous_version, current_version]
        for rule in self._get_version_increment_rules(stage_name):
            logger.info(
                f"Validating rule {rule.name!r} for {versions}"
            )
            validated = rule.validate_comparison(versions)
            if not validated:
                logger.error(
                    f"Validation failed for rule  {rule.name!r}: {rule.description}"
                )
                return validated
        return validated

    def _run_manifest_versions_comparison_rules(self, stage_name=None):
        validated = True
        for rule in self._get_manifest_versions_comparison_rules(stage_name):
            if not rule.manifest_names:
                continue
            manifests = [
                m for m in
                self._get_manifests(stage_name)
                if m.name in rule.manifest_names
            ]
            if not manifests:
                raise ValueError(f"Manifests {rule.manifest_names} not found.")
            versions = [m.version for m in manifests]
            logger.info(
                f"Validating rule {rule.name!r} for {versions}"
            )
            validated = rule.validate_comparison(versions)
            if not validated:
                logger.error(
                    f"Validation failed for rule {rule.name!r}: {rule.description}"
                )
                return validated
        return validated

    def _get_manifests(self, stage_name=None):
        manifests = self._manifests.copy()
        if stage_name:
            stage = self._get_stage(stage_name)
            manifests.extend(stage.manifests)
        return list(dict.fromkeys(manifests).keys())
    
    def _get_current_version_rules(self, stage_name=None) -> list[SingleVersionRule]:
        rules = self._current_version_rules.copy()
        if stage_name:
            stage = self._get_stage(stage_name)
            rules.extend(stage.current_version_rules)
        return list(dict.fromkeys(rules).keys())
    
    def _get_version_increment_rules(self, stage_name=None) -> list[VersionComparisonRule]:
        rules = self._version_increment_rules.copy()
        if stage_name:
            stage = self._get_stage(stage_name)
            rules.extend(stage.version_increment_rules)
        return list(dict.fromkeys(rules).keys())

    def _get_manifest_versions_comparison_rules(self, stage_name=None) -> list[ManifestsComparisonRule]:
        rules = self._manifest_versions_comparison_rules.copy()
        if stage_name:
            stage = self._get_stage(stage_name)
            rules.extend(stage.manifest_versions_comparison_rules)
        return list(dict.fromkeys(rules).keys())

    def _get_stage(self, stage_name):
        for stage in self._stages:
            if stage.name == stage_name:
                return stage
        raise ValueError(f"Stage {stage_name} not found.")



from logging import getLogger

from vertagus.core.project import Project
from vertagus.core.tag_base import Tag, AliasBase
from vertagus.core.scm_base import ScmBase


logger = getLogger(__name__)


def validate_project_version(scm: ScmBase,
                             project: Project,
                             stage_name: str = None
                             ) -> bool:
    previous_version = scm.get_highest_version()
    result = project.validate_version(
        previous_version,
        stage_name
    )
    current_version = project.get_version()
    if result:
        logger.info(f"Successfully validated current version: {current_version}")
    return result


def create_tags(scm: ScmBase,
                project: Project,
                stage_name: str = None,
                ref: str = None
                ) -> None:
    tag = Tag(project.get_version())
    scm.create_tag(tag, ref=ref)
    aliases = project.get_aliases(stage_name)
    for alias in aliases:
        scm.migrate_alias(alias, ref=ref)


def create_aliases(scm: ScmBase,
                   project: Project,
                   stage_name: str = None,
                   ref: str = None
                   ) -> None:
    aliases = project.get_aliases(stage_name)
    for alias in aliases:
        scm.migrate_alias(alias, ref=ref)



import os
from pathlib import Path
import sys

import click

from vertagus.configuration import load
from vertagus.configuration import types as cfgtypes
from vertagus import factory
from vertagus import operations as ops

_cwd = Path(os.getcwd())


def _try_get_config_path_in_cwd():
    if "vertagus.toml" in os.listdir(_cwd):
        return str(_cwd / "vertagus.toml")
    elif "vertagus.yml" in os.listdir(_cwd):
        return str(_cwd / "vertagus.yml")
    elif "vertagus.yaml" in os.listdir(_cwd):
        return str(_cwd / "vertagus.yaml")
    else:
        return None


@click.command("validate")
@click.option(
    "--config", 
    "-c", 
    default=None, 
    help="Path to the configuration file"
)
@click.option(
    "--stage-name",
    "-s",
    default=None,
    help="Name of a stage"
)
def validate_cmd(config, stage_name):
    if not config:
        config = _try_get_config_path_in_cwd()
    master_config = load.load_config(config)
    scm = factory.create_scm(
        cfgtypes.ScmData(**master_config["scm"])
    )
    default_package_root = Path(config).parent
    if "root" not in master_config["project"]:
        master_config["project"]["root"] = default_package_root
    project = factory.create_project(
        cfgtypes.ProjectData.from_project_config(master_config["project"])
    )
    if not ops.validate_project_version(
        scm=scm,
        project=project,
        stage_name=stage_name
    ):
        sys.exit(1)



from packaging import version
from vertagus.core.rule_bases import VersionComparisonRule


class Increasing(VersionComparisonRule):
    name = "any_increment"
    description = "Version must be greater than the previous one."

    def validate_comparison(self, versions: tuple[str, str]):
        version1, version2 = versions
        if not version1 and bool(version2):
            return True
        return version.parse(version1) < version.parse(version2)


class ManifestsComparisonRule(VersionComparisonRule):
    name = "manifests_comparison"
    description = "All manifests must have the same version."

    def __init__(self, config: dict):
        self.manifest_names = config["manifests"]

    def validate_comparison(self, versions: tuple[str, str]):
        if not versions:
            raise ValueError("No versions to compare.")
        if len(versions) == 1:
            raise ValueError("Only one version to compare. To compare, provide at least two versions.")
        return all([v == versions[0] for v in versions])



import pytest
from unittest.mock import Mock, patch
from vertagus.operations import validate_project_version, create_tags
from vertagus.core.project import Project
from vertagus.core.scm_base import ScmBase
from vertagus.core.tag_base import Tag


def test_validate_project_version():
    mock_scm = Mock(spec=ScmBase)
    mock_scm.get_highest_version.return_value = 'v1.0.0'
 
    mock_project = Mock(spec=Project)
    mock_project.validate_version.return_value = True
    mock_project.get_version.return_value = 'v1.0.1'
 
    with patch('vertagus.operations.logger') as mock_logger:
        result = validate_project_version(mock_scm, mock_project, 'stage1')
        mock_logger.info.assert_called_once_with("Successfully validated current version: v1.0.1")
    assert result is True


def test_create_tags_normal():
    mock_scm = Mock(spec=ScmBase)
    mock_project = Mock(spec=Project)
    mock_project.get_version.return_value = 'v1.0.1'
    mock_project.get_aliases.return_value = ['alias1', 'alias2']
    mock_scm.create_tag.return_value = None
    mock_scm.migrate_alias.return_value = None

    create_tags(mock_scm, mock_project, 'stage1', 'ref1')

    mock_scm.create_tag.assert_called_once()
    mock_scm.migrate_alias.assert_any_call('alias1', ref='ref1')
    mock_scm.migrate_alias.assert_any_call('alias2', ref='ref1')


def test_create_tags_no_stage():
    mock_scm = Mock(spec=ScmBase)
    mock_project = Mock(spec=Project)
    mock_project.get_version.return_value = 'v1.0.1'
    mock_project.get_aliases.return_value = []  
    mock_scm.create_tag.return_value = None
    create_tags(mock_scm, mock_project, ref='ref1')
    mock_scm.create_tag.assert_called_once()
    mock_scm.migrate_alias.assert_not_called()



from unittest.mock import MagicMock
import pytest
import os
from vertagus import factory
from vertagus.core.rule_bases import SingleVersionRule
from vertagus.configuration import types as t
from vertagus.core.scm_base import ScmBase


class DummyManifest:
    def __init__(self, **kwargs):
        pass


    
@pytest.fixture
def mock_manifest_cls(monkeypatch):
    mock = MagicMock(return_value=DummyManifest)
    monkeypatch.setattr(factory, "get_manifest_cls", mock)
    return mock

def test_create_manifests(mock_manifest_cls):
    manifest_data_list = [t.ManifestData(name="test_manifest", type="dummy_type", path="test_path", loc=["1", "2"]),]

    result = factory.create_manifests(manifest_data_list, "root_path")

    # We verify that the 'get_manifest_cls' function was called with the correct arguments
    mock_manifest_cls.assert_called_with("dummy_type")

    # We verify that a list of ManifestBase instances is returned
    assert len(result) == len(manifest_data_list)
    for item in result:
        assert isinstance(item, DummyManifest)


@pytest.fixture
def mock_single_version_rules(monkeypatch):
    mock_rule_getter = MagicMock()
    monkeypatch.setattr(factory, "get_single_version_rules", mock_rule_getter)
    return mock_rule_getter

def test_create_single_version_rules(mock_single_version_rules):
    rule_names = ["rule1", "rule2", "rule3"]

    result = factory.create_single_version_rules(rule_names)

    # Verify that the 'get_single_version_rules' function was called with the correct arguments
    mock_single_version_rules.assert_called_with(rule_names)

    # We verify that a list of SingleVersionRule instances is returned
    for item in result:
        assert isinstance(item, SingleVersionRule)


@pytest.fixture
def mock_scm_data():
    return t.ScmData(root="root", type="dummy_type", **{"key": "value"})

@pytest.fixture
def mock_scm_cls():
    class MockScm(ScmBase):
        def __init__(self,
                    root: str = None,
                    tag_prefix: str = None,
                    user_data: dict = None,
                    remote_name: str = None,
                    **kwargs
                    ):
            self.root = root or os.getcwd()
            self.tag_prefix = tag_prefix
            self.user_data = user_data 
            self.remote_name = remote_name

        def create_tag(self, tag, ref: str=None):
            pass
        
        def delete_tag(self, tag_name: str):
            pass
        
        def list_tags(self, prefix: str=None):
            return ["tag1", "tag2", "tag3"]

        def get_highest_version(self, prefix: str=None):
            pass

        def migrate_alias(self, alias: str, ref: str = None):
            pass

    return MockScm

def test_create_scm(mock_scm_data, mock_scm_cls, monkeypatch):
    monkeypatch.setattr(factory, "get_scm_cls", MagicMock(return_value=mock_scm_cls))
    scm = factory.create_scm(mock_scm_data)
    factory.get_scm_cls.assert_called_with("dummy_type")
    assert scm.root == "root"
    assert scm.tag_prefix is None


@pytest.fixture
def project_data():
    return t.ProjectData(
        manifests=[t.ManifestData(name="test_manifest", type="dummy_type", path="test_path", loc=["1", "2"])],
        rules=t.RulesData(current=["rule1", "rule2"], increment=["rule3"]),
        stages=[
            t.StageData(
                name="stage1",
                manifests=[t.ManifestData(name="test_manifest", type="dummy_type", path="test_path", loc=["1", "2"])],
                rules=t.RulesData(current=["rule1", "rule2"], increment=["rule3"])
            )],
        aliases=["alias1", "alias2"]
    )




from dataclasses import asdict
import os.path

from vertagus.core.project import Project
from vertagus.core.stage import Stage
from vertagus.core.tag_base import AliasBase
from vertagus.core.manifest_base import ManifestBase
from vertagus.core.rule_bases import SingleVersionRule, VersionComparisonRule
from vertagus.core.scm_base import ScmBase

from vertagus.providers.scm.registry import get_scm_cls
from vertagus.providers.manifest.registry import get_manifest_cls
from vertagus.aliases.loader import get_aliases
from vertagus.rules.single_version.loader import get_rules as get_single_version_rules
from vertagus.rules.comparison.loader import get_rules as get_version_comparison_rules

from .configuration import types as t


def create_project(data: t.ProjectData) -> Project:
    return Project(
        manifests=create_manifests(data.manifests, data.root),
        current_version_rules=create_single_version_rules(data.rules.current),
        version_increment_rules=create_version_comparison_rules(data.rules.increment, {}),
        manifest_versions_comparison_rules=create_version_comparison_rules(
            ["manifests_comparison"],
            {"manifests": data.rules.manifest_comparisons}
        ) if data.rules.manifest_comparisons else [],
        stages=create_stages(data.stages, data.root) if data.stages else None,
        aliases=create_aliases(data.aliases)
    )


def create_manifests(manifest_data: list[t.ManifestData], root: str = None) -> list[ManifestBase]:
    manifests = []
    for each in manifest_data:
        if root:
            each.path = os.path.join(root, each.path)
        manifest_cls = get_manifest_cls(each.type)
        manifests.append(manifest_cls(**each.config()))
    return manifests


def create_single_version_rules(rule_names: list[str]) -> list[SingleVersionRule]:
    return get_single_version_rules(rule_names)


def create_version_comparison_rules(rule_names: list[str], config) -> list[VersionComparisonRule]:
    rule_classes = get_version_comparison_rules(rule_names)
    return [rule_cls(config=config) for rule_cls in rule_classes]


def create_aliases(alias_names: list[str]) -> list[AliasBase]:
    return get_aliases(alias_names)


def create_stages(stage_data: list[t.StageData], project_root: str = None) -> list[Stage]:
    stages = []
    for data in stage_data:
        stages.append(Stage(
            name=data.name,
            manifests=create_manifests(data.manifests, project_root),
            current_version_rules=create_single_version_rules(data.rules.current),
            version_increment_rules=create_version_comparison_rules(data.rules.increment, {}),
            manifest_versions_comparison_rules=create_version_comparison_rules(
                ["manifests_comparison"],
                {"manifests": data.rules.manifest_comparisons}
            ) if data.rules.manifest_comparisons else [],
            aliases=create_aliases(data.aliases),
        ))
    return stages


def create_scm(data: t.ScmData) -> ScmBase:
    scm_cls = get_scm_cls(data.scm_type)
    return scm_cls(**data.config())



import typing as t
from click.formatting import (
    HelpFormatter,
    measure_table,
    iter_rows,
    wrap_text as click_wrap_text
)
from click import style


def wrap_text(text: str,
              width: int,
              initial_indent: str = "",
              subsequent_indent: str = "",
              preserve_paragraphs: bool = False
              ) -> str:
    text = click_wrap_text(
        text,
        width,
        initial_indent=initial_indent,
        subsequent_indent=subsequent_indent,
        preserve_paragraphs=preserve_paragraphs
    )
    lines = []
    for line in text.splitlines():
        if len(line) < width:
            line = line.ljust(width)
        lines.append(line)
    return "\n".join(lines)


def set_col_widths(col_widths: list[int], width: int):
    total_width = sum(col_widths)
    if total_width > width:
        col_widths = [int(width * col_width / total_width) for col_width in col_widths]
    return col_widths


class DisplayTableFormatter(HelpFormatter):
    
    def write_table(self, 
                    rows: t.Sequence[t.Tuple],
                    col_widths: t.Sequence[int],
                    col_spacing: int = 2,
                    header: bool = False
                    ) -> None:
        rows = list(rows)
        col_widths = set_col_widths(col_widths, self.width)
        for row_idx, cells in enumerate(iter_rows(rows, len(rows[0]))):
            is_header = header and row_idx == 0
            row_height = 1
            for idx, cell in enumerate(cells):
                if idx == 0:
                    cell_text = wrap_text(
                        cell,
                        col_widths[idx],
                        " " * self.current_indent,
                        subsequent_indent=" " * self.current_indent
                    )
                else:
                    indent_len = 0
                    for prev_cell_idx in range(0, idx):
                        indent_len += col_widths[prev_cell_idx] + col_spacing
                    cell_text = wrap_text(
                        cell,
                        col_widths[idx],
                        " " * col_spacing,
                        subsequent_indent=" " * indent_len
                    )
                cell_height = cell_text.count("\n") + 1
                row_height = max(row_height, cell_height)
                if cell_height < row_height:
                    cell_text += "\n" * (row_height - cell_height)
                if is_header:
                    cell_text = style(cell_text, bold=True)
                self.write(cell_text)
            self.write("\n")
        self.write("\n")



import typing as T
from vertagus.core.manifest_base import ManifestBase
from vertagus.core.rule_bases import SingleVersionRule, VersionComparisonRule
from vertagus.rules.comparison.library import ManifestsComparisonRule
from vertagus.core.tag_base import AliasBase
from .package_base import Package


class Stage(Package):
    
    def __init__(self,
                 name: str,
                 manifests: list[ManifestBase],
                 current_version_rules: list[T.Type[SingleVersionRule]],
                 version_increment_rules: list[VersionComparisonRule],
                 manifest_versions_comparison_rules: list[ManifestsComparisonRule],
                 aliases: list[type[AliasBase]] = None
                 ):
        super().__init__(
            manifests=manifests,
            current_version_rules=current_version_rules,
            version_increment_rules=version_increment_rules,
            manifest_versions_comparison_rules=manifest_versions_comparison_rules
        )
        self.name = name
        self.aliases = aliases or []

    @property
    def current_version_rules(self):
        return self._current_version_rules
    
    @property
    def version_increment_rules(self):
        return self._version_increment_rules
    
    @property
    def manifest_versions_comparison_rules(self):
        return self._manifest_versions_comparison_rules
    
    @property
    def manifests(self):
        return self._manifests

    def get_version_aliases(self, version: str) -> list[AliasBase]:
        return [alias(version) for alias in self.aliases]
